package comparator.multiObjective;

import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import problem.Problem;
import solution.MOSolution;
import solution.Solution;
import utility.BadBoundsException;
import utility.Utils;

public class HypeComparator extends GenericMOComparator
{
    
    //the upper bounds for the objective function values
    private double[] boundsVec ;
    //the lower bounds for the objective function values
    private double[] lowerBoundsVec ;
    private int largestDim;
    private int smallestDim;
    private int sampleSize = -1; // -1 means use exact calculation, other numbers are for sampling
    private int noObjectives;
    private boolean boundsInitialised = false;
    
    @Override
    public void initialise(Properties props, Problem problem)
    {
	sampleSize = Integer.parseInt(props.getProperty("HV_sample"));
	lowerBoundsVec = new double[ problem.getObjectives() ];
	noObjectives = problem.getObjectives();
	for (int i = 0; i < problem.getObjectives(); i++)
	{
	    lowerBoundsVec[i] = Integer.MAX_VALUE;
	}
    }
    
    
    private void initialiseHVBoundsBasedOnPopulation(List <Solution> population)
    {
	Solution sol;
	boundsVec = new double[ noObjectives ];
	largestDim = 0;
	smallestDim = 0;
	for (Iterator <Solution> iterator = population.iterator(); iterator.hasNext();)
	{
	    sol = iterator.next();
	    for (int i = 0; i < noObjectives; i++)
	    {
		if (boundsVec[i] < ((MOSolution)sol).getFitnessForObjective(i))
		{
		    boundsVec[i] = ((MOSolution)sol).getFitnessForObjective(i)*1.2; //20% margin
		    if (boundsVec[i] > boundsVec[largestDim])
			largestDim = i;
		}
		if (lowerBoundsVec[i] > ((MOSolution)sol).getFitnessForObjective(i))
		{
		    lowerBoundsVec[i] = ((MOSolution)sol).getFitnessForObjective(i)*0.8; //20% margin
		    if (lowerBoundsVec[i] < lowerBoundsVec[smallestDim])
			smallestDim = i;
		}
	    }
	}
	boundsInitialised = true;
    }

    @Override
    public void prepare(List<Solution> population)
    {
	if (!boundsInitialised)
	    this.initialiseHVBoundsBasedOnPopulation(population);
	boolean done = false;
	while(!done){
	    try{
		double[] hypervolumeForPopulation = new double[ population.size() ];
		hypeIndicator( hypervolumeForPopulation, sampleSize, population.size(),population );
	    	for (int i = 0; i < hypervolumeForPopulation.length; i++){
	    	   ((MOSolution)population.get(i)).setHypervolume(hypervolumeForPopulation[i]);
		}
		done = true;
	    } catch (BadBoundsException e){
		initialiseHVBoundsBasedOnPopulation(population);
	    }
	}
	prepared = true;

    }

    @Override
    public int compare(Solution arg0, Solution arg1){
	if (!prepared){
	    System.err.println("Comparator unprepared!");
	    System.exit(1);
	}
	
	if ( ((MOSolution)arg0).getHypervolume() > ((MOSolution)arg1).getHypervolume())
	    return -1;
	else if(((MOSolution)arg0).getHypervolume() < ((MOSolution)arg1).getHypervolume())
	    return 1;
	return 0;
    }
    



    private void hypeIndicator( double[] val, int nrOfSamples, int param_k, List<Solution> pop ) throws BadBoundsException
        /**
         * Determine the hypeIndicator
         * \f[ \sum_{i=1}^k \left( \prod_{j=1}^{i-1} \frac{k-j}{|P|-j} \right) \frac{ Leb( H_i(a) ) }{ i } \f]
         *
         * if nrOfSamples < 0, then do exact calculation, else sample the indicator
         *
         * @param[out] val vector of all indicator values
         * @param[in] popsize size of the population \f$ |P| \f$
         * @param[in] lowerbound scalar denoting the lower vertex of the sampling box
         * @param[in] upperbound scalar denoting the upper vertex of the sampling box
         * @param[in] nrOfSamples the total number of samples or, if negative, flag
         * 		that exact calculation should be used.
         * @param[in] param_k the variable \f$ k \f$ "In the case of mating selection, k equals |P| and
         * in the case of environmental selection, k is the number of solutions to be removed from P.
         * @param[in] points matrix of all objective values dim*popsize entries
         * @param[in] rho weight coefficients
         */
        {
    	int i,j;
    	double[] rho = new double[param_k+1];
    	/** Set alpha */
    	rho[0] = 0;
    	for( i = 1; i <= param_k; i++ ){
    	    rho[i] = 1.0 / (double)i;
    	    for( j = 1; j <= i-1; j++ )
    		rho[i] *= (double)(param_k - j ) / (double)( val.length - j );
    	}
    	for( i = 0; i < val.length; i++ )
    	    val[i] = 0.0;
    	if( nrOfSamples < 0 )
    	{
    	    hypeExact( val, val.length,/* lowerbound, upperbound, */param_k, pop,
    		    rho );
    	}
    	else
    	{
    	    hypeSampling( val, val.length, /*lowerbound, upperbound,*/ nrOfSamples,
    		    param_k, pop, rho );
    	}
        }


        
       private void hypeSampling( double[] val, int popsize,/* double lowerbound,
    	    double upperbound,*/ int nrOfSamples, int param_k, List<Solution> pop,
    	    double[] rho ) throws BadBoundsException
        /**
         * Sampling the hypeIndicator
         * \f[ \sum_{i=1}^k \left( \prod_{j=1}^{i-1} \frac{k-j}{|P|-j} \right) \frac{ Leb( H_i(a) ) }{ i } \f]
         *
         * @param[out] val vector of all indicators
         * @param[in] popsize size of the population \f$ |P| \f$
         * @param[in] lowerbound scalar denoting the lower vertex of the sampling box
         * @param[in] upperbound scalar denoting the upper vertex of the sampling box
         * @param[in] nrOfSamples the total number of samples
         * @param[in] param_k the variable \f$ k \f$
         * @param[in] points matrix of all objective values dim*popsize entries
         * @param[in] rho weight coefficients
         * @pre popsize >= 0 && lowerbound <= upperbound && param_k >= 1 &&
         * 		param_k <= popsize
         */
        {
    	assert(popsize >= 0 );
    	//assert( lowerbound <= upperbound );
    	
    	assert( param_k >= 1 );
    	assert( param_k <= popsize );

    	int i, s, k;
    	int[] hitstat = new int[ popsize ];
    	int domCount;

    	double[] sample = new double[noObjectives];
    	for( s = 0; s < nrOfSamples; s++ )
    	{
    	    for( k = 0; k < noObjectives; k++ )
    		sample[ k ] = Utils.getRandomDoubleforInterval(lowerBoundsVec[k], boundsVec[k] );
    	    

    	    domCount = 0;
    	    for( i = 0; i < popsize; i++ )
    	    {
    		Solution sol =  pop.get(i);
    		if( weaklyDominates( ((MOSolution)sol).getAllFitnesses(), sample) )
    		{
    		    domCount++;
    		    if( domCount > param_k )
    			break;
    		    hitstat[i] = 1;
    		}
    		else
    		    hitstat[i] = 0;
    	    }
    	    if( domCount > 0 && domCount <= param_k )
    	    {
    		for( i = 0; i < popsize; i++ )
    		    if( hitstat[i] == 1 )
    			val[i] += rho[domCount];
    		
    	    }
    	}
    	for( i = 0; i < popsize; i++ )
    	{
    	    val[i] = val[i] * Math.pow((boundsVec[largestDim]-lowerBoundsVec[smallestDim]), noObjectives ) / (double)nrOfSamples;
    	    if (val[i] == 0)
    		throw new BadBoundsException();
    	}
        }
       

       void hypeExact( double[] val, int popsize,/* double lowerbound, double upperbound,*/
       		int param_k, List<Solution> pop, double[] rho  ) throws BadBoundsException
       //!!!popsize might be different from complete population size (might be front size) - but it probably isn't.
       /**
        * Calculating the hypeIndicator
        * \f[ \sum_{i=1}^k \left( \prod_{j=1}^{i-1} \frac{k-j}{|P|-j} \right) \frac{ Leb( H_i(a) ) }{ i } \f]
        */
       {
       	int i;
       	
       	int[] indices = new int[ popsize ];
       	
       	
       	for( i = 0; i < popsize; i++  )
       		indices[i] = i;

       	int [][] points = new int[popsize][noObjectives];
       	for (i = 0; i < points.length; i++)
    	{
    	   Solution sol = pop.get(i);
    	   System.arraycopy(((MOSolution)sol).getAllFitnesses(), 0, points[i], 0, noObjectives);
    	}
       	
       	/** Recursively calculate the indicator values */
       	hypeExactRecursive( points, popsize, popsize, noObjectives-1, 
       			indices, val, rho, param_k );
       }

       
       private boolean weaklyDominates( double[] point1, double[] point2)
       {
       	boolean better = true;
       	int i = 0;


       	while( i < noObjectives && better )
       	{
       		better = point1[i] <= point2[i];
       		i++;
       	}
       	return better;
       }


       void hypeExactRecursive( int [][] input_p, int pnts, int nrOfPnts,
    	   int actDim, int[] input_pvec, double[] fitness,
    	   double[] rho, int param_k ) throws BadBoundsException
       /**
        * Internal function used by hypeExact
        */
       {
           int i, j;
           double extrusion;
           int[] pvec = new int[pnts];
           int[][] p = new int[pnts][noObjectives];
           for( i = 0; i < pnts; i++ ) {
    	   fitness[i] = 0;
    	   pvec[i] = input_pvec[i];
           }
           for( i = 0; i < pnts; i++ )
           {
    	   p[i] = input_p[i];
           }
    	  

           rearrangeIndicesByColumn( p, nrOfPnts, noObjectives, actDim, pvec );

           for( i = 0; i < nrOfPnts; i++ )
           {
    	   if( i < nrOfPnts - 1 )
    	       extrusion = p[pvec[i+1]][actDim ] - p[pvec[i]][actDim ];
    	   else
    	       extrusion = boundsVec[actDim] - p[ pvec[i]][actDim ];
    	   
    	   if (extrusion < 0)//this means the reference point is wrong and the calculations will fail
    	       throw new BadBoundsException();

    	   if( actDim == 0 ) {
    	       if( i+1 <= param_k )
    		   for( j = 0; j <= i; j++ ) {
    		       fitness[ pvec[j] ] = fitness[ pvec[j] ]
    		                                     + extrusion*rho[ i+1 ];
    		   }
    	   }
    	   else if( extrusion > 0 ) {
    	       double[] tmpfit = new double[pnts];
    	       //check the effect of tmpfit
    	       hypeExactRecursive( p, pnts, i+1, actDim-1, pvec,
    		       tmpfit, rho, param_k );
    	       for( j = 0; j < pnts; j++ )
    		   fitness[j] += extrusion*tmpfit[j];
    	   }
           }
       }
       
       void rearrangeIndicesByColumn(int [][] mat, int rows, int columns, int col,
    	   int [] ind )
       /**
        * Internal function used by hypeExact
        */
       {
           int  maxLevels =  300;
           int[]  beg = new int[maxLevels]; int[] end = new int[maxLevels];
           int i = 0, L, R, swap;
           double pref;
           int pind;
           double[] ref = new double[rows];
           //ind is list of indices of size of population 
           //rows is number of points
           //columns is number of objectives
           //col is actDim
           for( i = 0; i < rows; i++ ) {
    	   ref[i] = mat[ind[i]][col];//should be equivalent of below!
    	  // ref[i] = mat[ col + ind[i]*columns ];
           }
           i = 0;

           beg[0] = 0; end[0] = rows;
           while ( i >= 0 ) {
    	   L = beg[i]; R = end[i]-1;
    	   if( L < R ) {
    	       pref = ref[L];
    	       pind = ind[L];
    	       while( L < R ) {
    		   while( ref[R] >= pref && L < R )
    		       R--;
    		   if( L < R ) {
    		       ref[L] = ref[R];
    		       ind[ L++] = ind[R];
    		   }
    		   while( ref[L] <= pref && L < R )
    		       L++;
    		   if( L < R) {
    		       ref[ R ] = ref[ L ];
    		       ind[ R--] = ind[L];
    		   }
    	       }
    	       ref[ L ] = pref; ind[L] = pind;
    	       beg[i+1] = L+1; end[i+1] = end[i];
    	       end[i++] = L;
    	       if( end[i] - beg[i] > end[i-1] - beg[i-1] ) {
    		   swap = beg[i]; beg[i] = beg[i-1]; beg[i-1] = swap;
    		   swap = end[i]; end[i] = end[i-1]; end[i-1] = swap;
    	       }
    	   }
    	   else {
    	       i--;
    	   }
           }
      }

    //use this once for each problem to add a field to the problem file.
    // The field should be of the format: HVReferencePoints = {<boundsVec[0],boundsVec[1],> etc}
    // It should be inserted between the metadata and the first matrix on its own line
       public void showBoundsVec()
       {
	   System.out.println("Bounds ");
		for (int i = 0; i< boundsVec.length ; i++)
		{
		    boundsVec[i] *= 0.8;
		    boundsVec[i] *= 1.5;
		    System.out.println(boundsVec[i]+" ");
		}
       }

}
