
package comparator.multiObjective;

import java.util.List;
import java.util.Properties;
import problem.Problem;
import solution.MOSolution;
import solution.Solution;
import utility.Utils;

/**
 *
 * @author Alda
 */
public class MSOPSComparator extends GenericMOComparator{
    private int targets; /* number of target vectors to use */
    private int q = 100;		 /* scaling factor in VADS. It defines the 'sharpness' of the probe.
     q=10 is blunt, q=50 is not so blunt,  q=100 is usually a good choice and is sharp enough to find a reasonable
     objective surface, without promoting too many bad solutions. */
    private boolean vads = true;	 /* dual optimization switch  false (No VADS) true (+VADS) */
    private double[][] targetVectors;
    private int[][] MSOPSRanks; //Dual Optimization Matrix
    private int nObjectives;

    @Override
    public void initialise(Properties props, Problem problem){
        nObjectives = problem.getObjectives();
	targets = Integer.parseInt(props.getProperty("targets"));
        if(targets<nObjectives){
            System.out.println("target should be >="+nObjectives);
            targets=nObjectives;
        }
       // initialise vectors randomly. TODO: This should probably be done differently
        targetVectors = new double[targets][nObjectives];
        for(int i=0; i<targets;i++){
            for(int j=0; j<nObjectives; j++){
                targetVectors[i][j]=Utils.random.nextDouble();
            }
        }
    }

    @Override
    public void prepare(List<Solution> population ){
        popSize = population.size();
        if(vads)
            MSOPSRanks = new int[popSize][2*targets];
        else
            MSOPSRanks = new int[popSize][targets]; //Dual Optimization Matrix
	double[][] minmaxScores = new double[popSize][targets]; // Score Matrix n x t of Weighted MinMax
	double[][] vadsScores = new double[popSize][targets]; // Score Matrix n x t of VADS
	for(int i = 0; i < population.size(); i++){ // compute scores
	 // for each solution
            Solution sol = population.get(i);
            for(int j = 0; j < targets; j++) { // and each weight vector               
                minmaxScores[i][j]=weightedMinMax((MOSolution)sol, targetVectors[j]); // compute scores
                MSOPSRanks[i][j]=1;
                if(vads){
                    vadsScores[i][j]= vectorAngleDistanceScaling((MOSolution)sol, targetVectors[j]);
                    MSOPSRanks[i][j+targets] = 1;
                }
            }
	}
	int factor = vads ? 2 : 1;
        
	for(int i = 0; i < popSize; i++){ // take each individual
            Solution sol = population.get(i);
            for(int j = (i + 1); j < popSize; j++){	// compare it to every individual with heigher index
                for(int k = 0; k < targets; k++){ // for every weight vector
                    if(minmaxScores[i][k] > minmaxScores[j][k]){// check which scoreMinmax is higher and increment rank
                        MSOPSRanks[i][factor*k]+=1;
                    }
                    else if(minmaxScores[i][k] < minmaxScores[j][k]){
                        MSOPSRanks[j][factor*k]+=1;
                    }
                    if(vads){
                        if(vadsScores[i][k] > vadsScores[j][k])// check which vadsScore is higher and increment rank
                            MSOPSRanks[i][factor*k+1]+=1;
                        else if(vadsScores[i][k] < vadsScores[j][k])
                            MSOPSRanks[j][factor*k+1] +=1;
                    }
                }
            }
            
            sortRanks(MSOPSRanks[i]);// sort ranks ascending
            // and write it into the MSOPSRanks array
            int size_f;
            if(vads) size_f = 2 * targets;
            else size_f = targets;     
            for(int k = 0; k < size_f; k++)
                ((MOSolution)sol).setMSOPSRank(MSOPSRanks[i][k], k, targets);
	}
    }

    private void sortRanks(int[] solRanks){
        for(int i=0; i<solRanks.length;i++){
            for(int j=i+1; j<solRanks.length-1;j++){
                if(solRanks[i]>solRanks[j]){
                    int temp = solRanks[i];
                    solRanks[i] = solRanks[j];
                    solRanks[j]= temp;
                }
            }
        }
    }

    private double weightedMinMax(MOSolution individual, double[] weights){
        double maxValue = -1;
        double actValue;
        for(int i = 0; i < nObjectives; i++){
            actValue = individual.getFitnessForObjective(i) * weights[i];
            if(actValue > maxValue) maxValue = actValue;
        }
        return maxValue;
    }

    /**
    * Calculates the VADS score given:
    * @param a multiobjective solution MOGASolution sol
    * @param a vector of weights double[] weights
    * @return double VADS score
    */
    private double vectorAngleDistanceScaling(MOSolution sol, double[] weights){
	double score = 0;
	double[] solWeight = new double[nObjectives];
	double weightMagnitude = 0;
	double[] solObjectives = new double[nObjectives];
	double objectivesMagnitude = 0;

	// Calculate magnitude of vectors
	for(int i = 0; i < nObjectives; i++){
            weightMagnitude = weightMagnitude + Math.pow(1/weights[i],2);
            objectivesMagnitude = objectivesMagnitude + Math.pow(sol.getFitnessForObjective(i),2);
	}
	weightMagnitude = Math.sqrt(weightMagnitude);
	objectivesMagnitude = Math.sqrt(objectivesMagnitude);
	
	//Normalize vectors and calculate dot product of weight- and objective-vector
	for(int i = 0; i < nObjectives; i++){
            solWeight[i] = (1/weights[i]) / weightMagnitude;
            solObjectives[i] = sol.getFitnessForObjective(i) / objectivesMagnitude;
            score = score + solWeight[i] * solObjectives[i];
	}
	//Calculate VADS score
	score = objectivesMagnitude / Math.pow(score, q);
 	return score;
    }

    @Override
    public int compare(Solution sola, Solution solb){
	int i, target_size;
	if(vads) target_size = 2* targets;
	else target_size = targets;
	for(i = 0; i < target_size; i++){
            if(((MOSolution)sola).getMSOPSRank(i) > ((MOSolution)solb).getMSOPSRank(i))
                return -1;            
            else if (((MOSolution)sola).getMSOPSRank(i) < ((MOSolution)solb).getMSOPSRank(i))
                return 1;
	}
	return 0;
    }
}
