package solver;

import comparator.constraints.ConstrainedGenericComparator;
import solution.QAP.ConstrainedQAPSolution;
import java.util.List;
import solution.Solution;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Properties;

import java.util.BitSet;

import problem.Problem;
import utility.OutOfFunctionEvaluationsException;
import utility.StagnationException;
import utility.Utils;
import utility.reporting.Reporting;

public class ConstrainedGA extends Algorithm
{
    private double mutationRate;
    private LinkedList<Solution> matingPool;
    protected List<Solution> population;
    protected int populationSize;

    @Override
    public void initialise(Properties props, Problem problem)
    {
	super.initialise(props, problem);
        ((ConstrainedGenericComparator)currentComparator).initialise(props, problem);
	noImprovementCounter = 0;
	bestFitness = 0;
	population = new LinkedList<Solution>();
	matingPool = new LinkedList<Solution>();
       populationSize = Integer.parseInt(props.getProperty("population_size"));

	for (int i = 0; i < populationSize; i++)
	{
            try{
                Solution sol = (Solution) Class.forName(props.getProperty("solution")).newInstance();
                sol.initialiseRandomly(problem);
                sol.evaluateSolution(problem);
                population.add(sol);
            } catch(Exception ex){
                break;
            }    
	}
    }

    @Override
    public void changeParameters(Properties props){
        populationSize = Integer.parseInt(props.getProperty("population_size"));
	matingPoolSize = Integer.parseInt(props.getProperty("mating_pool"));
        mutationRate = Double.parseDouble(props.getProperty("mutation_prob"));
    }
    private void addToPopulationNoDuplicates(Solution sol){
	for (int i = population.size()-1; i > 0 ; i--){
	    if (population.get(i).getCurrentFitness() == sol.getCurrentFitness())
		return;
	}
	population.add(sol);
    }
  
    @Override
    public void oneIteration(Reporting reporting) {
        //doStochasticUniversalSelection();
        //doBinaryTournamentWithNiching();
        doBinaryTournament();
        Solution sol1, sol2;
        int x1, x2;

        while(!matingPool.isEmpty()){
            x1 = Utils.random.nextInt(matingPool.size());
            sol1 = matingPool.remove(x1);
            if (matingPool.size()> 1){
                do{
                    x2 = Utils.random.nextInt(matingPool.size());
                }while (x1 == x2);
                sol2 = (ConstrainedQAPSolution) matingPool.remove(x2);
                try{
                    doCrossover(sol1, sol2);
                }catch (OutOfFunctionEvaluationsException e){
                    doSelection();
                }
            }
            try{
                doMutation(sol1);
            }
            catch (OutOfFunctionEvaluationsException e){
                doSelection();
            }
        }
        doSelection();
        try{
            checkImprovement(reporting);
        } catch (StagnationException e){
            try {
                reinitialisePopulation();
            } catch (OutOfFunctionEvaluationsException ex) {
               doSelection();
            }
        }
        
    }
    private void reinitialisePopulation() throws OutOfFunctionEvaluationsException{
	for (Iterator<Solution> iterator = population.iterator(); iterator.hasNext();){
	    Solution sol = iterator.next();
	    sol.initialiseRandomly(optimisationProblem);
	    sol.evaluateSolution(optimisationProblem);
	}

    }
    
    private void doStochasticUniversalSelection(){
	double max = population.get(0).getPenaltyFitness();
	double min = population.get(population.size()-1).getPenaltyFitness();
	double interval = (max - min)/(matingPoolSize+1);
	double currentLocation = population.get(0).getPenaltyFitness() -  Utils.random.nextDouble()*interval;
	int i = 0;
	int j = 0;
	while ( i < matingPoolSize){
	    currentLocation -= interval;
	    j = 0;
	    while (j < population.size()-1){
		if (currentLocation<population.get(j).getPenaltyFitness() && currentLocation > population.get(j+1).getPenaltyFitness()){
		    if (population.get(j).getPenaltyFitness() - currentLocation < currentLocation - population.get(j+1).getPenaltyFitness())
			matingPool.add(population.get(j));
		    else
			matingPool.add(population.get(j+1));
		    break;
		}
		else
		    j++;
	    }

	    i++;
	}
    }
    
    private void doCrossover(Solution sol1, Solution sol2) throws OutOfFunctionEvaluationsException{
	if (sol1 == sol2)
	    return;
	Solution [] children = sol1.uniformCrossover(sol2);
	children[0].evaluateSolution(optimisationProblem);
	children[1].evaluateSolution(optimisationProblem);
	/*System.out.println("Parent 1 "+sol1);
	System.out.println("Parent 2 "+sol2);
	System.out.println("Child 1 "+children[0]);
	System.out.println("Child 2 "+children[1]);*/
	this.addToPopulationNoDuplicates(children[0]);
	this.addToPopulationNoDuplicates(children[1]);
    }

    private void doMutation(Solution sol) throws OutOfFunctionEvaluationsException{
	Solution sol2 = sol.uniformSwapMutatation(mutationRate);
	sol2.evaluateSolution(optimisationProblem);
	this.addToPopulationNoDuplicates(sol2);
    }
    
    
    protected void doSelection()
    {
	((ConstrainedGenericComparator)currentComparator).prepare(population);
	Collections.sort(population, currentComparator);
	((ConstrainedGenericComparator)currentComparator).unprepare();
	
	for (int i = population.size()-1; i >= populationSize; i--)
	    population.remove(i);
	
    }

    private void checkImprovement(Reporting reporting) throws StagnationException
    {
	if (population.get(0).getCurrentFitness() == bestFitness)
	{
	    noImprovementCounter++;
	}
	else
	{
	    bestFitness = population.get(0).getCurrentFitness();
	    noImprovementCounter = 0;
	}
	if (noImprovementCounter == 50)
	{
	    reporting.addBest(population.get(0).clone());
	    throw new StagnationException();
	}
    }

    @Override
    public List<Solution> getPopulation() {
        return population;
    }



   private void doBinaryTournament()
    {
	Solution sol1, sol2;
	int x1, x2 ;
	BitSet bs = new BitSet();//get rid of duplication
	for (int i = 0; i < matingPoolSize; i++)
	{
	    do{
		x1 = Utils.random.nextInt(populationSize);
		sol1 =  population.get(x1);
	    }while (bs.get(x1)) ;
	    do{
		x2 = Utils.random.nextInt(populationSize);
		sol2 = population.get(x2);
	    }while (bs.get(x2) || x1 == x2 );

	    int result = currentComparator.compare(sol1, sol2);
	    if (result==-1)
	    {
		matingPool.add(sol1);
		bs.set(x1);
	    }
	    else
	    {
		matingPool.add(sol2);
		bs.set(x2);
	    }

	}

    }
    /* This niching method is down to Deb (An investigation of niche and species formation in genetic function optimisation, 1989)
     * well, he did it in a continuous environment with the 0.1 factor; The timeout of 0.25*N is authentic*/
 /*   private void doBinaryTournamentWithNiching()
    {
	Solution sol1, sol2 = null;
	int x1, x2 ;
	int timeout = populationSize/4;
	int timeoutCounter = 0;
	double distance = 0;
	for (int i = 0; i < matingPoolSize; i++)
	{
	    timeoutCounter = 0;
	    x1 = Utils.random.nextInt(populationSize);
	    sol1 =  population.get(x1);
	    //System.out.println("sol1 "+sol1.currentFitness);
	    //System.out.println("sol2 "+sol2.currentFitness);

	    do{
		timeoutCounter++;
		if (timeoutCounter > timeout)
		{
		    matingPool.add(sol1);
		    sol1 = null;
		    // System.out.println("Breaking ties ");
		    break;
		}
		x2 = Utils.random.nextInt(populationSize);
		sol2 = population.get(x2);
		distance = Utils.getNormalisedHammingDistance(sol1, sol2, problem.getCardinality());
		//System.out.println("sol1 "+ x1+ " "+sol1.currentFitness);
		//System.out.println("sol2 "+ x2+ " "+sol2.currentFitness);
		//System.out.println("Distance "+distance);
	    }while (distance < 0.3 || x1 == x2 );

	    if (sol1 == null)
		continue;
	    if (sol1.currentFitness > sol2.currentFitness)
	    {
		matingPool.add(sol1);

	    }
	    else
	    {
		matingPool.add(sol2);

	    }

	}

    }*/




}
