package solver.multiobjective;

import java.util.LinkedList;
import java.util.Properties;

import problem.Problem;
import problem.QAP.MOQAPProblem;
import solution.MOSolution;
import utility.OutOfFunctionEvaluationsException;
import utility.Utils;
import utility.reporting.MOReporting;
import utility.reporting.Reporting;

/* HV is based on minimisation - better HVs mean lower objective function values*/
public abstract class MOGA extends MOAlgorithm
{

    private int mutationDistance;
    private int tournamentSize;


    private MOQAPProblem MOProblem;



    @Override
    public void initialise(Properties props, Problem problem)
    {
	super.initialise(props, problem);
	MOProblem = (MOQAPProblem)problem;
	mutationDistance = Integer.parseInt(props.getProperty("mutation_dist"));
	tournamentSize = Integer.parseInt(props.getProperty("tournament_size"));
	MOPopulation = new LinkedList<MOSolution>();
	matingPool = new LinkedList<MOSolution>();


	for (int i = 0; i < populationSize; i++)
	{
	    MOSolution sol = (MOSolution) population.get(i);
            try{
                sol.evaluateSolution(problem);
            }catch(OutOfFunctionEvaluationsException ex){
                break;
            }
	    MOPopulation.add(sol);
	}

    }




    @Override
    public void run(Reporting reporting){
        while(MOProblem.availableFunctionEvaluations()){
            while(true){
                try{
                    createMatingPool();
                    MOSolution sol1, sol2;
                    int x1, x2;
                    while(!matingPool.isEmpty()){
                        x1 = Utils.random.nextInt(matingPool.size());
                        sol1 = (MOSolution) matingPool.remove(x1);
                        if (matingPool.size()> 1){
                            do{
                                x2 = Utils.random.nextInt(matingPool.size());
                            }while (x1 == x2);
                            sol2 = (MOSolution) matingPool.remove(x2);
                            doDistancePreservingCrossover(sol1, sol2);
                        }
                        //doMultipointMutation(sol1, mutationProb);
                        doDistancePreservingMutation(sol1, mutationDistance);
                    }
                    doSelection();
                } catch (OutOfFunctionEvaluationsException e){
                    finishOff(reporting);
                }
            }

        }
    }

    private void doDistancePreservingCrossover(MOSolution sol1,
	    MOSolution sol2) throws OutOfFunctionEvaluationsException{
	MOSolution sol = sol1.distancePreservingCrossover(sol2);
	sol.evaluateSolution(MOProblem);
	MOPopulation.add(sol);
    }

    private void doDistancePreservingMutation(MOSolution sol1, int distance) throws OutOfFunctionEvaluationsException{
	MOSolution sol = sol1.distancePreservingMutation(distance);
	sol.evaluateSolution(MOProblem);
	MOPopulation.add(sol);
    }

    protected abstract void createMatingPool();

    protected MOSolution doTournamentSelection(){
	if (tournamentSize == 2)
	    return doBinaryTournament(); //faster??
	MOSolution[] sols = new MOSolution[tournamentSize];
	int index = Utils.random.nextInt(MOPopulation.size());
	boolean theSame = false;
	sols[0] = MOPopulation.get(index);
	for (int i = 1; i < tournamentSize; i++){
	    theSame = false;
	    do{
		index = Utils.random.nextInt(MOPopulation.size());
		for (int j = 0; j < i; j++){
		    if (sols[index] == sols[j]){
			theSame = true;
			break;
		    }
		}
	    }while (theSame);
	    sols[i] = MOPopulation.get(index);
	}
	MOSolution largestHV = sols[0];

	for (int i = 1; i < tournamentSize; i++){
	    if (currentMOComparator.compare(largestHV, sols[i])==1)
		largestHV = sols[i];
	}
	return largestHV;
    }

    private MOSolution doBinaryTournament()
    {
	MOSolution sol1, sol2 = null;
	int x1, x2;

	x1 = Utils.random.nextInt(MOPopulation.size());
	sol1 = (MOSolution) MOPopulation.get(x1);
	do{
	    x2 = Utils.random.nextInt(MOPopulation.size());
	}while (x1 == x2); //problem - why population size of 1?
	sol2 = (MOSolution) MOPopulation.get(x2);
	if (currentMOComparator.compare(sol1, sol2)==-1)
	    return(sol1);
	else
	    return(sol2);

    }


    protected abstract void doSelection();




   private void finishOff(Reporting reporting)
   {
        fastExtractNonDomFronts();
	((MOReporting)reporting).addToBest(nondominatedFronts.get(0));
   }


}



